------------------------------ MODULE Streams -------------------------------
EXTENDS NaturalsInduction, WellFoundedInduction, Functions, SequenceTheorems

(***************************************************************************)
(* We model a stream (or infinite sequence) in canonical fashion as a      *)
(* function whose domain is the set of positive integers. (We let the      *)
(* indexes of a stream start at 1 for compatibility with finite sequences. *)
(* In particular, the finite prefix of an infinite sequence seq can be     *)
(* written as Restrict(seq, 1..n) and is a finite sequence.                *)
(* Infinite sequences are used in the statement and proof of determinism   *)
(* of PharOS where we show that the prefix of execution is compatible in a *)
(* precise sense with some prefix of a statically chosen infinite system   *)
(* execution.                                                              *)
(***************************************************************************)

Natp == {n \in Nat : n >= 1}

Stream(S) == [Natp -> S]

LEMMA Natp_induction ==
  ASSUME NEW P(_), P(1),
         ASSUME NEW k \in Natp, P(k) PROVE P(k+1)
  PROVE  \A n \in Natp : P(n)
<1>. DEFINE Q(k) == P(k+1)
<1>1. Q(0)
  OBVIOUS
<1>2. ASSUME NEW k \in Nat, Q(k)
      PROVE  Q(k+1)
  BY <1>2 DEF Natp
<1>3. \A k \in Nat : Q(k)
  BY <1>1, <1>2, NatInduction
<1>. QED  BY <1>3, Isa DEF Natp

LEMMA Natp_wellfounded == IsWellFoundedOn(OpToRel(<,Natp), Natp)
  <1>1. /\ Natp \subseteq Nat
        /\ OpToRel(<,Natp) \cap (Nat \X Nat) \subseteq OpToRel(<,Nat)
    BY DEF Natp, OpToRel
  <1>2. IsWellFoundedOn(OpToRel(<,Natp), Nat)
    BY <1>1, IsWellFoundedOnSubrelation, NatLessThanWellFounded
  <1>. QED
    BY <1>1, <1>2, IsWellFoundedOnSubset

LEMMA RestrictStream ==
  ASSUME NEW S, NEW w \in Stream(S), NEW n \in Nat
  PROVE  /\ Restrict(w, 1..n) \in Seq(S)
         /\ Len(Restrict(w, 1..n)) = n
BY DEF Restrict, Stream, Natp

LEMMA RestrictStreamMonotonic ==
  ASSUME NEW S, NEW w \in Stream(S), 
         NEW m \in Nat, NEW n \in Nat, m <= n
  PROVE  IsPrefix(Restrict(w, 1..m), Restrict(w, 1..n))
<1>. /\ Restrict(w, 1..m) \in Seq(S)
     /\ Restrict(w, 1..n) \in Seq(S)
     /\ DOMAIN Restrict(w, 1..m) = 1..m
  BY DEF Restrict, Stream, Natp
<1>1. Len(Restrict(w, 1..m)) <= Len(Restrict(w, 1..n))
  BY RestrictStream
<1>2. Restrict(w, 1..m) = Restrict(Restrict(w, 1..n), 1..m)
  BY DEF Restrict, Stream, Natp
<1>. QED  BY <1>1, <1>2, IsPrefixProperties

LEMMA IsPrefixRestrictStream ==
  ASSUME NEW S, NEW w \in Stream(S), NEW n \in Nat,
         NEW s \in Seq(S), IsPrefix(s, Restrict(w, 1..n))
  PROVE  \E m \in 0 .. n : s = Restrict(w, 1..m)
<1>1. /\ Len(s) <= n
      /\ s = Restrict(Restrict(w, 1..n), DOMAIN s)
  BY IsPrefixProperties, RestrictStream
<1>2. /\ Restrict(w, 1 .. Len(s)) \in Seq(S)
      /\ Len(Restrict(w, 1 .. Len(s))) = Len(s)
  BY RestrictStream
<1>25. /\ DOMAIN s = 1 .. Len(s)
       /\ 1 .. Len(s) \subseteq 1 .. n
  BY <1>1
<1>3. \A i \in 1 .. Len(s) : s[i] = Restrict(w, 1 .. Len(s))[i]
  BY <1>1, <1>25, Zenon DEF Restrict
<1>. QED  BY <1>1, <1>2, <1>3, SeqEqual

=============================================================================
\* Modification History
\* Last modified Wed Jul 15 14:33:51 CEST 2015 by merz
\* Created Mon Jul 13 09:22:36 CEST 2015 by merz
