------------------------------- MODULE System_Spec -------------------------------
EXTENDS Types, Filters, SequenceTheorems

USE DEF Time  \** doesn't seem to be inherited from Types (?)
(***************************************************************************)
(* This module contains the specification of PharOS executions, at a high  *)
(* level of abstraction, as well as proofs of elementary invariants.       *)
(*                                                                         *)
(* The current configuration of the system is represented as follows:      *)
(* - a function associating its current state to each agent,               *)
(* - the set of all messages sent so far, and                              *)
(* - the current system time.                                              *)
(* The specification also records the complete system history. The history *)
(* variable is not used for specifying the behavior of the system, but for *)
(* stating the property of determinism.                                    *)
(*                                                                         *)
(* System steps arise from some agent executing its next instruction,      *)
(* within the time bounds associated with the instruction. The system time *)
(* advances, respecting the deadlines of all agents: we assume that the    *)
(* system blocks as soon as some deadline is violated. (The proof of       *)
(* witness existence in module Witness shows that our assumptions imply    *)
(* that some non-blocking execution exists.)                               *)
(***************************************************************************)

VARIABLES state, (*  state is what is called Y_i in the paper *)
          messages, (* message register *)
          time, (* current time *)
          history (* history is an extension of what is called \mathbf{Y} in the paper *)

vars == <<state, messages, time, history>>
USE DEF vars

State_type ==
  /\ state \in [Agent -> State]
  /\ messages \in SUBSET Message
  /\ time \in Time

History_type ==
  /\ history \in Seq(SystemState)
  /\ history # << >>

Init ==
  /\ state = initState
  /\ messages = {} 
  /\ time = 1
  /\ history = << [st |-> state, msgs |-> messages, t |-> time ] >>

Next == 
  /\ time' \in {t \in Time : t >= time}
  /\ \A a \in Agent : time' <= instr(state[a]).late
  /\ \E a \in Agent :
        /\ instr(state[a]).early <= time'
        /\ LET res == exec(state[a], messages)
           IN  /\ state' = [state EXCEPT ![a] = res[1]]
               /\ messages' = res[2]
  /\ history' = Append(history, [st |-> state', msgs |-> messages', t |-> time'])

Spec == Init /\ [][Next]_vars

(* Type correctness of the specification *)
LEMMA Spec_type ==  Spec => []State_type /\ []History_type
<1>1 Init => State_type /\ History_type
  BY initState_type DEF Init, State_type, History_type, SystemState
<1>2 State_type /\ History_type /\ [Next]_vars
     => State_type' /\ History_type'
  BY exec_type, AppendProperties DEF Next, State_type, History_type, SystemState
<1> QED  BY PTL, <1>1, <1>2 DEF Spec

Inv == 
  /\ Last(history) = [st |-> state, msgs |-> messages, t |-> time]
  /\ \A i \in 1 .. Len(history) : history[i].t <= time
  /\ \A a \in Agent : time <= instr(state[a]).late

LEMMA Spec_Inv == Spec => []Inv
<1>1. Init => Inv
  BY initState_good, GoodSystemState_late DEF Init, Last, Inv
<1>2. Inv /\ State_type /\ State_type' /\ History_type /\ Next => Inv'
  <2>. SUFFICES ASSUME Inv, State_type, State_type', History_type,
                       time' \in Time, time <= time',
                       \A a \in Agent : time' <= instr(state[a]).late,
                       NEW a \in Agent,
                       state' = [state EXCEPT ![a] = exec(state[a], messages)[1]],
                       messages' = exec(state[a], messages)[2],
                       history' = Append(history, [st |-> state', msgs |-> messages', t |-> time'])
                PROVE  Inv'
    BY DEF Next
  <2>1. Last(history') = [st |-> state', msgs |-> messages', t |-> time']
    BY AppendProperties DEF History_type, Last
  <2>2. ASSUME NEW i \in 1 .. Len(history')
        PROVE  history'[i].t <= time'
    <3>. time \in Time  BY DEF State_type
    <3>1. CASE i \in 1 .. Len(history)
      BY <3>1, Z3 DEF Inv, History_type, SystemState
    <3>2. CASE i = Len(history)+1
      BY <3>2 DEF History_type
    <3>. QED  BY <3>1, <3>2
  <2>3. ASSUME NEW b \in Agent
        PROVE  time' <= instr(state'[b]).late
    BY instr_type, exec_monotonic DEF State_type, DatedInstruction
  <2>. QED  BY <2>1, <2>2, <2>3, Zenon DEF Inv
<1>3. Inv /\ UNCHANGED vars => Inv'
  BY DEF Inv
<1> QED BY <1>1, <1>2, <1>3, PTL, Spec_type DEF Spec

(***************************************************************************)
(* All messages sent during the execution were generated from some send    *)
(* instruction corresponding to a state recorded in the history, and all   *)
(* such instructions whose deadlines have passed were actually sent.       *)
(***************************************************************************)
(**
Inv_messages == 
  LET instrOf(a) == { instr(history[i].st[a]) : i \in DOMAIN history }
      sends(a) == { di \in instrOf(a) : di.ins.type = "send" }
      toSend == UNION { sends(a) : a \in Agent }
      expired == { m \in toSend : m.late < time }
  IN  /\ messages \subseteq toSend
      /\ expired \subseteq messages
**)
Inv_messages == 
  LET states(a) == { history[i].st[a] : i \in 1 .. Len(history) }
      toSend == UNION { sent(states(a)) : a \in Agent }
      expired == { m \in toSend : m.late < time }
  IN  /\ messages \subseteq toSend
      /\ expired \subseteq messages

LEMMA Spec_inv_messages == Spec => []Inv_messages
<1>. DEFINE states(a) == { history[i].st[a] : i \in 1 .. Len(history) }
            toSend == UNION { sent(states(a)) : a \in Agent }
            expired == { m \in toSend : m.late < time }
            IInv == /\ messages \subseteq toSend
                    /\ \A a \in Agent : sent(states(a)) \subseteq messages \cup {instr(state[a])}
<1>1. Init => IInv
  BY DEF Init, sent
<1>2. IInv /\ State_type /\ History_type /\ Inv /\ [Next]_vars => IInv'
  <2>. SUFFICES ASSUME IInv,
                       State_type, History_type, Inv,
                       time' \in Time, time' >= time,
                       \A b \in Agent : time' <= instr(state[b]).late,
                       NEW a \in Agent,
                       instr(state[a]).early <= time',
                       state' = [state EXCEPT ![a] = exec(state[a], messages)[1]],
                       messages' = exec(state[a], messages)[2],
                       history' = Append(history, [st |-> state', msgs |-> messages', t |-> time'])
                PROVE  IInv'
    BY Isa DEF Next
  <2>1. /\ Len(history) \in Nat \ {0}
        /\ Len(history') = Len(history)+1
        /\ \A k \in 1 .. Len(history) : history'[k] = history[k]
        /\ history'[Len(history)+1].st = state'
    BY AppendProperties, LenProperties DEF History_type
  <2>2. messages' \subseteq toSend'
    <3>1. CASE instr(state[a]).ins.type = "send"
      <4>1. messages' = messages \union {instr(state[a])}
        BY <2>1, <3>1 DEF exec
      <4>4. instr(state[a]) \in toSend
        BY <2>1, <3>1, instr_type DEF Inv, State_type, Last, sent, Message
      <4>5. toSend \subseteq toSend'
        BY <2>1 DEF History_type, sent
      <4>. QED  BY Zenon, <4>1, <4>4, <4>5
    <3>2. CASE instr(state[a]).ins.type # "send"
      BY <2>1, <3>2, exec_msgs DEF State_type, History_type, sent
    <3>. QED  BY Zenon, <3>1, <3>2
  <2>3. ASSUME NEW b \in Agent, NEW di \in sent(states(b))', di \notin messages'
        PROVE  di = instr(state'[b])
    <3>2. PICK i \in 1 .. Len(history') : di = instr(history'[i].st[b]) /\ di \in Message
      BY <2>3 DEF sent
    <3>3. CASE i \in 1 .. Len(history)
      <4>1. di \in sent(states(b))
        BY <3>2, <3>3, <2>1 DEF sent
      <4>2. di \notin messages
        BY <2>1, <2>3, exec_msgs DEF State_type
      <4>3. di = instr(state[b])
        BY <4>1, <4>2, Zenon
      <4>4. b = a => di \in messages'
        BY <4>3, <3>2, Zenon DEF exec, Message
      <4>. QED  BY <4>3, <4>4, <2>1, <2>3 DEF State_type
    <3>4. CASE i = Len(history)+1  BY <3>2, <3>4, <2>1
    <3>. QED  BY <3>2, <3>3, <3>4, <2>1
  <2>. QED  BY Zenon, <2>2, <2>3
<1>3. IInv /\ State_type /\ Inv => Inv_messages
  BY instr_type DEF State_type, Inv, Inv_messages, DatedInstruction, sent
<1>. QED  BY <1>1, <1>2, <1>3, Spec_type, Spec_Inv, PTL DEF Spec

=============================================================================
\* Modification History
\* Last modified Tue Jan 26 11:47:45 CET 2016 by merz
\* Last modified Fri Nov 21 11:39:59 CET 2014 by shaolin
\* Last modified Thu Aug 28 15:36:01 CEST 2014 by jael
\* Last modified Thu Aug 28 11:06:04 CEST 2014 by doligez
\* Last modified Tue Jul 01 17:35:34 CEST 2014 by selma
\* Created Thu Feb 13 10:35:01 CET 2014 by selma
