-------------------------------- MODULE Types -------------------------------
EXTENDS Sequences, Streams, TLAPS, TLC

CONSTANT
  Agent,      (* set of agents *) 
  State,      (* set of agents' states *)
  initState,  (* initial assignment of states to agents *)
  MsgId,      (* set of exchanged messages *)
  instr(_),   (* operator to access the pending instruction of a state *) 
  visible(_,_), (* visibility condition between dated instructions (including messages) *)
  Update_fun,   (* functions mapping states to states, used by instructions *)
  Update      (* set of functions updating local states: component of instructions *)

Time == Nat
USE DEF Time
State_Domain == State (* auxiliary operator, useful for overriding when experimenting with TLC *)

(***************************************************************************)
(* Basic assumptions about agents and the initial states.                  *)
(***************************************************************************)
ASSUME Agent_nempty == Agent # {}
ASSUME initState_type == initState \in [Agent -> State]

LEMMA State_nempty == State # {}
BY Agent_nempty, initState_type

(***************************************************************************)
(* We consider three types of instructions that agents execute:            *)
(* - update :  corresponds to a local action where the state of the agent  *)
(*             is updated through an Update function;                      *)
(* - send :    sends a messages in addition to updating the local state;   *)
(* - receive : attempts to receive a message and updates the local state,  *)
(*             the update function depends on whether the message could be *)
(*             received (i.e. was previously sent and is visible, cf.      *)
(*             operator msg_received defined below).                       *)
(***************************************************************************)
Instruction ==
         [ type : {"update"}, upd : Update ]
  \union [ type : {"send"}, msg : MsgId, upd : Update ]
  \union [ type : {"receive"}, msg : MsgId, bupd : [BOOLEAN -> Update] ]

(***************************************************************************)
(* A dated instruction consists of an instruction and a time window for    *)
(* executing this instruction.                                             *)
(***************************************************************************)
DatedInstruction ==
  { x \in [ins : Instruction, early : Time, late : Time] : x.early < x.late }

(***************************************************************************)
(* The ``next'' function picks the dated instruction that the process      *)
(* wants to execute in a given system state.                               *)
(***************************************************************************)
ASSUME instr_type == \A s \in State : instr(s) \in DatedInstruction

(***************************************************************************)
(* Messages are dated instructions of type "send".                         *)
(***************************************************************************)
Message == { di \in DatedInstruction : di.ins.type = "send" }

(***************************************************************************)
(* Set of all messages that correspond to a set of states.                 *)
(***************************************************************************)
sent(states) == { instr(s) : s \in states } \cap Message

(***************************************************************************)
(* A system state is given by an array of agent states, the history of     *)
(* all messages sent so far, and the current time.                         *)
(***************************************************************************)
SystemState == [st: [Agent -> State], msgs: SUBSET Message, t: Time]

(***************************************************************************)
(* Given a system state s and agent a that executes, time should advance   *)
(* to some point t such that all of the following conditions hold:         *)
(* - t is at least as big as the time at system state s,                   *)
(* - agent a is executable at time t, and                                  *)
(* - no agent misses its deadline by time t.                               *)
(***************************************************************************)
GoodNextTime(s,a,t) ==
  /\ s.t <= t
  /\ instr(s.st[a]).early <= t
  /\ \A b \in Agent : t <= instr(s.st[b]).late

(***************************************************************************)
(* A ``good'' system state is one at which some agent can execute and for  *)
(* which a next time instant exists that respects the above constraints.   *)
(***************************************************************************)
GoodSystemState ==
  { s \in SystemState : \E a \in Agent, t \in Time : GoodNextTime(s,a,t) }

LEMMA GoodSystemState_late == 
  ASSUME NEW s \in GoodSystemState, NEW a \in Agent
  PROVE s.t <= instr(s.st[a]).late
<1> /\ s.t \in Time
    /\ instr(s.st[a]).late \in Time
  BY instr_type DEF GoodSystemState, SystemState,  DatedInstruction
<1> PICK b \in Agent, t \in Time: GoodNextTime(s,b,t)
  BY DEF GoodSystemState
<1> s.t <= t
  BY DEF GoodNextTime
<1> t <= instr(s.st[a]).late
  BY DEF GoodNextTime
<1> QED  BY DEF Time

(***************************************************************************)
(* The initial system state is a ``good'' state. In fact, all system       *)
(* states encountered during an execution will be seen to be good.         *)
(***************************************************************************)
ASSUME initState_good ==
  [ st |-> initState, msgs |-> {}, t |-> 1 ] \in GoodSystemState

(***************************************************************************)
(* Update functions map states to states and can only increase the start   *)
(* and end dates of the instructions that will be executed.                *)
(***************************************************************************)
ASSUME Update_type == Update \in SUBSET [State -> State]
ASSUME Update_monotonic == \A state \in State, upd \in Update :
  /\ instr(state).early <= instr(upd[state]).early
  /\ instr(state).late <= instr(upd[state]).late

(***************************************************************************)
(* Assumption about the visibility condition: a dated instruction di can   *)
(* be visible for another dated instruction dj only if the end date of di  *)
(* is strictly smaller than the starting date of dj, i.e. if dj cannot     *)
(* occur before di in any system execution.                                *)
(***************************************************************************)
ASSUME visible_cond == \A di,dj \in DatedInstruction : 
  /\ visible(di,dj) \in BOOLEAN
  /\ visible(di,dj) => di.late < dj.early

msgReceived(msgs, n) == \E i \in msgs : n.ins.msg = i.ins.msg /\ visible(i,n)

(***************************************************************************)
(* The following operator models the execution of a process given its      *)
(* current state and the current set of messages. It returns a pair        *)
(* consisting of the state after executing the instruction that the        *)
(* process wants to execute and the new set of messages.                   *)
(***************************************************************************)
exec(state, msgs) ==
  LET i == instr(state)
  IN  IF i.ins.type = "update"
      THEN << i.ins.upd[state], msgs >>
      ELSE IF i.ins.type = "send"
      THEN << i.ins.upd[state], msgs \union {i} >>
      ELSE (* i.ins.type = "receive" *)
           << i.ins.bupd[msgReceived(msgs, i)][state], msgs >>

LEMMA exec_type ==
  ASSUME NEW state \in State, NEW msgs \in SUBSET Message
  PROVE  /\ exec(state, msgs) \in State \X (SUBSET Message)
         /\ \E f \in Update : exec(state, msgs)[1] = f[state]
<1>. DEFINE i == instr(state)
<1>. i \in DatedInstruction
  BY instr_type
<1>1. CASE i.ins \in [ type : {"update"}, upd : Update ]
  BY <1>1, Update_type DEF exec
<1>2. CASE i.ins \in [ type : {"send"}, msg : MsgId, upd : Update ]
  BY <1>2, Update_type DEF exec, Message
<1>3. CASE i.ins \in [ type : {"receive"}, msg : MsgId, bupd : [BOOLEAN -> Update] ]
  <2>1. exec(state, msgs) \in State \X (SUBSET Message)
    BY <1>3, Update_type, visible_cond, Zenon DEF exec, msgReceived
  <2>2. exec(state, msgs)[1] = i.ins.bupd[msgReceived(msgs, i)][state]
    BY <1>3 DEF exec
  <2>3. i.ins.bupd[msgReceived(msgs, i)] \in Update
    BY <1>3, Zenon DEF msgReceived
  <2>5. \E f \in Update : exec(state, msgs)[1] = f[state]
    BY <2>2, <2>3
  <2>. QED  BY <2>1, <2>5
<1>. QED
  BY <1>1, <1>2, <1>3, Zenon DEF DatedInstruction, Instruction

LEMMA exec_msgs ==
  ASSUME NEW state \in State, NEW msgs \in SUBSET Message
  PROVE  /\ msgs \subseteq exec(state, msgs)[2]
         /\ instr(state).ins.type # "send" => exec(state, msgs)[2] = msgs
BY DEF exec

(***************************************************************************)
(* The successor state from a given state only depends on the set of       *)
(* messages that have been received.                                       *)
(***************************************************************************)
LEMMA exec_from_same_state ==
  ASSUME NEW s \in State, NEW msgs1, NEW msgs2,
         instr(s).ins.type = "receive" => 
              (msgReceived(msgs1, instr(s)) <=> msgReceived(msgs2, instr(s)))
  PROVE  exec(s, msgs1)[1] = exec(s, msgs2)[1]
BY instr_type, Zenon DEF Instruction, DatedInstruction, msgReceived, exec
  
    
LEMMA exec_monotonic == 
  ASSUME NEW state \in State, NEW msgs \in SUBSET Message
  PROVE  /\ instr(state).early <= instr(exec(state,msgs)[1]).early
         /\ instr(state).late <= instr(exec(state,msgs)[1]).late
BY Update_monotonic, exec_type

(***************************************************************************)
(* The following predicate checks if the pair <<pre, post>> of (good)      *)
(* system states can be explained by a transition of agent a such that the *)
(* temporal conditions are respected.                                      *)
(***************************************************************************)
executes(a, pre, post) ==
  LET pstate == pre.st[a]
      res == exec(pstate, pre.msgs)
  IN  /\ GoodNextTime(pre, a, post.t)
      /\ post.st = [pre.st EXCEPT ![a] = res[1]]
      /\ post.msgs = res[2]

LEMMA executes_preserves_GoodSystemState ==
  ASSUME NEW a \in Agent, 
         NEW pre \in GoodSystemState, NEW post \in SystemState,
         executes(a, pre, post)
  PROVE  post \in GoodSystemState
<1>. USE DEF Time
<1>. /\ post.st \in [Agent -> State]
     /\ post.msgs \in SUBSET Message
     /\ post.t \in Time
     /\ \A b \in Agent : /\ instr(post.st[b]) \in DatedInstruction
                         /\ instr(post.st[b]).early \in Time
                         /\ instr(post.st[b]).late \in Time
  BY instr_type DEF SystemState, DatedInstruction
<1>. DEFINE P(d) == \E b \in Agent : instr(post.st[b]).early <= post.t + d
<1>. HIDE DEF P
<1>1. PICK d \in Nat : /\ P(d)
                       /\ \A k \in 0 .. d-1 : ~P(k)
  <2>1. PICK k \in Nat : instr(post.st[a]).early <= post.t + k
    BY Z3
  <2>2. P(k)
    BY <2>1 DEF P
  <2>. QED  BY <2>2, SmallestNatural, Isa
<1>2. PICK c \in Agent : instr(post.st[c]).early <= post.t + d
  BY <1>1 DEF P
<1>4. ASSUME NEW b \in Agent
      PROVE  instr(post.st[b]).late >= post.t + d
  <2>. SUFFICES ASSUME instr(post.st[b]).late < post.t + d
                PROVE  FALSE
    OBVIOUS
  <2>1. instr(post.st[b]).early < post.t + d
    BY DEF DatedInstruction
  <2>2. instr(post.st[b]).early <= post.t + (d-1)
    BY <2>1
  <2>3. d = 0
    BY <2>2, <1>1 DEF P
  <2>4. instr(post.st[b]).late >= post.t
    <3>. /\ pre.st \in [Agent -> State]
         /\ pre.msgs \in SUBSET Message
         /\ instr(pre.st[b]) \in DatedInstruction
         /\ instr(pre.st[b]).late \in Time
      BY instr_type DEF GoodSystemState, SystemState, DatedInstruction
    <3>1. post.t <= instr(pre.st[b]).late
      BY DEF executes, GoodNextTime
    <3>2. instr(pre.st[b]).late <= instr(post.st[b]).late
      <4>1. CASE b = a
        BY <4>1, exec_monotonic DEF executes
      <4>2. CASE b # a
        BY <4>2 DEF executes
      <4>. QED  BY <4>1, <4>2
    <3>. QED  BY <3>1, <3>2
  <2>. QED  BY <2>3, <2>4
<1>5. GoodNextTime(post, c, post.t + d)
  BY <1>2, <1>4 DEF GoodNextTime
<1>. QED  BY <1>5 DEF GoodSystemState

LEMMA executes_monotonic ==
  ASSUME NEW H \in [Natp -> SystemState],
         \A n \in Natp : \E a \in Agent : executes(a, H[n], H[n+1]),
         NEW i \in Natp, NEW j \in Natp, i <= j
  PROVE  H[i].t <= H[j].t
<1>. USE DEF Natp, Time
<1>. DEFINE P(k) == H[i].t <= H[i+k].t
<1>1. P(0)
  BY DEF SystemState
<1>2. ASSUME NEW k \in Nat, P(k)
      PROVE  P(k+1)
  BY <1>2 DEF executes, GoodNextTime, SystemState
<1>3. \A k \in Nat : P(k)
  BY <1>1, <1>2, NatInduction, Isa
<1>4. PICK k \in Nat : j = i+k
  BY Z3
<1>. QED  BY <1>3, <1>4, Zenon

(***************************************************************************)
(* Time progress assumption: in any infinite sequence of system states     *)
(* that correspond to executing agents, time should advance beyond any     *)
(* bound. Since agent execution relies on advancing the current time to    *)
(* some appropriate value, this assumption essentially requires that the   *)
(* lower bounds advance in any infinite execution.                         *)
(***************************************************************************)
ASSUME TimeProgress ==
  \A H \in [Natp -> GoodSystemState] :
     (\A n \in Natp : \E a \in Agent : executes(a, H[n], H[n+1]))
  => \A T \in Time : \E k \in Natp : H[k].t > T

(*
TakeSend(msgs) == { m \in msgs : m.ins.type = "send" }
ApplyNext(states) == { instr(s) : s \in states }
*)
=============================================================================
